package net.asukaze.cassava.core.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;

import net.asukaze.cassava.core.model.DataFormat;
import net.asukaze.cassava.core.model.ICassavaTableModel;
import net.asukaze.cassava.io.util.CsvReader;

public class DataFormatSetting {
	public static final DataFormatSetting INSTANCE = new DataFormatSetting();

	private static final String USER_FORMAT_FILE_NAME = "/format.csv"; //$NON-NLS-1$
	private static final String DEFAULT_FORMAT_FILE_NAME = "format.csv"; //$NON-NLS-1$
	private static final String IGNORE_KEYWORD = "#"; //$NON-NLS-1$

	private List<DataFormat> formats;

	private DataFormatSetting() {
		try {
			formats = new ArrayList<DataFormat>();
			read();
		} catch (IOException ex) {
			ex.printStackTrace();
		}
	}

	private String unescape(String escaped) {
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < escaped.length(); i++) {
			char c = escaped.charAt(i);
			if (c == '\\') {
				char n = escaped.charAt(++i);
				if (n == 'n') {
					sb.append('\n');
				} else if (n == 'r') {
					sb.append('\r');
				} else if (n == 't') {
					sb.append('\t');
				} else if (n == '_') {
					sb.append(' ');
				} else {
					sb.append(n);
				}
			} else {
				sb.append(c);
			}
		}
		return sb.toString();
	}

	private void read() throws IOException {
		InputStream is;
		is = DataFormatSetting.class.getResourceAsStream(USER_FORMAT_FILE_NAME);
		if (is == null) {
			is = DataFormatSetting.class
					.getResourceAsStream(DEFAULT_FORMAT_FILE_NAME);
		}
		CsvReader reader = new CsvReader(DataFormat.RFC4180,
				new InputStreamReader(is));
		ICassavaTableModel model;
		try {
			model = reader.readTable();
		} finally {
			reader.close();
		}
		for (int y = 0; y < model.getRowCount(); y++) {
			String name = model.getValueAt(y, 0);
			if (name.length() == 0 || name.startsWith(IGNORE_KEYWORD)) {
				continue;
			}
			boolean isDefault = Boolean.parseBoolean(model.getValueAt(y, 1));
			String[] exts = model.getValueAt(y, 2).split(";");
			String strongDelimiters = unescape(model.getValueAt(y, 3));
			String weakDelimiters = unescape(model.getValueAt(y, 4));
			String quoteChars = unescape(model.getValueAt(y, 5));
			boolean useQuote = Boolean.parseBoolean(model.getValueAt(y, 6));
			boolean omitComma = Boolean.parseBoolean(model.getValueAt(y, 7));
			DataFormat format = new DataFormat(name, isDefault, exts,
					strongDelimiters, weakDelimiters, quoteChars, useQuote,
					omitComma);
			formats.add(format);
		}
	}

	public DataFormat getDefault() {
		for (DataFormat format : formats) {
			if (format.isDefault()) {
				return format;
			}
		}
		if (formats.size() > 0) {
			return formats.get(0);
		}
		return null;
	}

	public DataFormat get(String name) {
		for (DataFormat format : formats) {
			if (format.getName().equals(name)) {
				return format;
			}
		}
		return null;

	}

	public DataFormat getDataFormatForExt(String ext) {
		for (DataFormat format : formats) {
			for (String x : format.getExts()) {
				if (x.equals(ext)) {
					return format;
				}
			}
		}
		return null;
	}

	private String getExt(String filename) {
		int index = filename.lastIndexOf('.');
		if (index >= 0) {
			return filename.substring(index + 1);
		} else {
			return null;
		}
	}

	public DataFormat getDataFormatForFilename(String filename) {
		String ext = getExt(filename);
		DataFormat format = null;
		if (ext != null) {
			format = getDataFormatForExt(ext);
			if (format != null) {
				return format;
			}
		}
		return getDefault();
	}

	public List<DataFormat> getAll() {
		return formats;
	}
}
