package net.asukaze.cassava.ui;

import java.awt.BorderLayout;
import java.awt.Image;
import java.awt.Toolkit;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JMenuBar;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JToolBar;
import javax.swing.UIManager;

import net.asukaze.cassava.core.command.Command;
import net.asukaze.cassava.core.command.CommandException;
import net.asukaze.cassava.core.model.CassavaTableModel;
import net.asukaze.cassava.io.command.OpenFileCommand;
import net.asukaze.cassava.ui.action.Action;
import net.asukaze.cassava.ui.action.RefreshAction;
import net.asukaze.cassava.ui.table.CassavaGrid;
import net.asukaze.cassava.ui.table.CassavaRowHeaderList;

public class MainFrame extends JFrame implements IApplication {
	private static final long serialVersionUID = 1L;
	private static final String[] ICON_FILES = {
			"icon/cassava16.gif", "icon/cassava32.gif" }; //$NON-NLS-1$ //$NON-NLS-2$

	private Map<Class<?>, Action> actions = new HashMap<Class<?>, Action>();
	private Action startUpAction = null;
	private JPanel jContentPane = null;
	private JPanel pnlHeader = null;
	private CassavaGrid cassavaGrid = null;
	private JPopupMenu popupMenu = null;
	private JMenuBar menuBar = null;
	private JToolBar toolBar = null;
	private JButton btnNew = null;
	private Properties properties = new Properties();

	public static void main(String[] args) {
		try {
			UIManager
					.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsLookAndFeel");
		} catch (Exception noWindows) {
			try {
				UIManager
						.setLookAndFeel("com.sun.java.swing.plaf.nimbus.NimbusLookAndFeel");
			} catch (Exception noNimbus) {
			}
		}

		MainFrame mainFrame = new MainFrame();
		mainFrame.setVisible(true);
		for (String arg : args) {
			try {
				mainFrame.openFile(arg);
			} catch (IOException ex) {
				Log.error(ex);
			}
		}
		mainFrame.getCassavaGrid().setValueAt("", 3, 3);
		if (mainFrame.getStartUpAction() != null) {
			try {
				mainFrame.getStartUpAction().run();
			} catch (Exception ex) {
				Log.error(ex);
			}
		}
	}

	/**
	 * This is the default constructor
	 */
	public MainFrame() {
		super();
		initialize();
	}

	/**
	 * This method initializes this
	 * 
	 * @return void
	 */
	@SuppressWarnings("serial")
	private void initialize() {
		this.setSize(500, 375);
		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		this.setContentPane(getJContentPane());
		this.setTitle(Messages.getString(Messages.CASSAVA_TITLE));

		this.setJMenuBar(getJMenuBar());
		MenuBuilder menuBuilder = new MenuBuilder();
		menuBuilder.readMenuCsv(this, getJMenuBar(), getPopupMenu());

		this.setTransferHandler(new FileDropHandler() {
			@Override
			public void fileDropped(File file) throws IOException {
				openFile(file.getAbsolutePath());
			}
		});

		Toolkit tk = this.getToolkit();
		List<Image> icons = new ArrayList<Image>();
		for (String file : ICON_FILES) {
			icons.add(tk.getImage(MainFrame.class.getResource(file)));
		}
		this.setIconImages(icons);
	}

	/**
	 * This method initializes jContentPane
	 * 
	 * @return javax.swing.JPanel
	 */
	private JPanel getJContentPane() {
		if (jContentPane == null) {
			jContentPane = new JPanel();
			jContentPane.setLayout(new BorderLayout());
			// jContentPane.add(getPnlHeader(), BorderLayout.NORTH);
			CassavaGrid grid = getCassavaGrid();
			JScrollPane scrollPane = new JScrollPane(grid);
			CassavaRowHeaderList headerList = new CassavaRowHeaderList(grid);
			headerList.setBackground(scrollPane.getViewport().getBackground());
			scrollPane.setRowHeaderView(headerList);
			jContentPane.add(scrollPane, BorderLayout.CENTER);
		}
		return jContentPane;
	}

	/**
	 * This method initializes pnlHeader
	 * 
	 * @return javax.swing.JPanel
	 */
	@SuppressWarnings("unused")
	private JPanel getPnlHeader() {
		if (pnlHeader == null) {
			pnlHeader = new JPanel();
			pnlHeader.setLayout(new BorderLayout());
			JToolBar toolBar = getToolBar();
			pnlHeader.setPreferredSize(toolBar.getPreferredSize());
			pnlHeader.add(toolBar, BorderLayout.CENTER);
		}
		return pnlHeader;
	}

	/**
	 * This method initializes cassavaGrid
	 * 
	 * @return javax.swing.JScrollPane
	 */
	public CassavaGrid getCassavaGrid() {
		if (cassavaGrid == null) {
			cassavaGrid = new CassavaGrid();
			cassavaGrid.setModel(new CassavaTableModel());
			cassavaGrid.setComponentPopupMenu(getPopupMenu());
		}
		return cassavaGrid;
	}

	public JPopupMenu getPopupMenu() {
		if (popupMenu == null) {
			popupMenu = new JPopupMenu();
		}
		return popupMenu;
	}

	/**
	 * This method initializes toolBar
	 * 
	 * @return javax.swing.JToolBar
	 */
	private JToolBar getToolBar() {
		if (toolBar == null) {
			toolBar = new JToolBar();
			toolBar.add(getBtnNew());
		}
		return toolBar;
	}

	/**
	 * This method initializes menu
	 * 
	 * @return javax.swing.JMenuBar
	 */
	public JMenuBar getJMenuBar() {
		if (menuBar == null) {
			menuBar = new JMenuBar();
		}
		return menuBar;
	}

	public void registerAction(Action action) {
		actions.put(action.getClass(), action);
	}

	public Action getAction(Class<?> clazz) {
		return actions.get(clazz);
	}

	public void setStartUpAction(Action startUpAction) {
		this.startUpAction = startUpAction;
	}

	public Action getStartUpAction() {
		return startUpAction;
	}

	public void exit() {
		this.setVisible(false);
		this.dispose();
	}

	public JFrame getFrame() {
		return this;
	}

	public Properties getProperties() {
		return properties;
	}

	/**
	 * This method initializes btnNew
	 * 
	 * @return javax.swing.JButton
	 */
	private JButton getBtnNew() {
		if (btnNew == null) {
			btnNew = new JButton();
		}
		return btnNew;
	}

	public void openFile(String filename) throws IOException {
		try {
			CassavaGrid grid = getCassavaGrid();
			Reader reader = new FileReader(filename);
			Command cmd = new OpenFileCommand(grid.getTableModel(), grid
					.getViewModel(), reader, filename);
			grid.getCommandStack().run(cmd);
			RefreshAction.execute(grid);
		} catch (CommandException ex) {
			Throwable cause = ex.getCause();
			if (cause instanceof IOException) {
				throw (IOException) cause;
			} else {
				throw new IOException(ex);
			}
		}
	}
}
