package net.asukaze.cassava.ui;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.Map;

import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.KeyStroke;

import net.asukaze.cassava.core.model.DataFormat;
import net.asukaze.cassava.core.model.ICassavaTableModel;
import net.asukaze.cassava.io.util.CsvReader;
import net.asukaze.cassava.ui.action.Action;
import net.asukaze.cassava.ui.action.ActionNotFoundAction;

public class MenuBuilder {
	private static final String USER_MENU_FILE_NAME = "/menu.csv"; //$NON-NLS-1$
	private static final String DEFAULT_MENU_FILE_NAME = "menu.csv"; //$NON-NLS-1$
	private static final String IGNORE_KEYWORD = "#"; //$NON-NLS-1$
	private static final String STARTUP_KEYWORD = "!startup"; //$NON-NLS-1$
	private static final String POPUP_KEYWORD = "!popup"; //$NON-NLS-1$
	private static final String SEPARATOR_KEYWORD = "-"; //$NON-NLS-1$
	private static final String SUBMENU_KEYWORD = ">"; //$NON-NLS-1$
	private static final String AMP_REGEX = "\\&"; //$NON-NLS-1$
	private static final String EMPTY = ""; //$NON-NLS-1$

	private Map<String, JMenu> menuMap = new HashMap<String, JMenu>();

	private JMenuBar jMenuBar;
	private JPopupMenu jPopupMenu;

	private int getMnemonic(String text) {
		int pos = text.indexOf('&');
		if (pos >= 0 && pos < text.length() - 1) {
			return text.charAt(pos + 1);
		}
		return 0;
	}

	private JMenu getJMenu(String category) {
		if (menuMap.containsKey(category)) {
			return menuMap.get(category);
		}
		JMenu jmenu = new JMenu();
		jmenu.setMnemonic(getMnemonic(category));
		jmenu.setText(category.replaceAll(AMP_REGEX, EMPTY));
		jMenuBar.add(jmenu);
		menuMap.put(category, jmenu);
		return jmenu;
	}

	private Action getAction(MainFrame mainFrame, String actionName) {
		try {
			Class<?> clazz = Class.forName(actionName);
			Object object = clazz.newInstance();
			if (object instanceof Action) {
				Action action = (Action) object;
				action.setApplication(mainFrame);
				mainFrame.registerAction(action);
				return action;
			}
		} catch (ClassNotFoundException | SecurityException
				| IllegalArgumentException | InstantiationException
				| IllegalAccessException e) {
			// ActionNotFoundAction
		}
		Action action = new ActionNotFoundAction(actionName);
		action.setApplication(mainFrame);
		return action;
	}

	private KeyStroke getKeyStroke(String str) {
		str = str.replace('+', ' ');
		str = str.replace("Ctrl", "ctrl"); //$NON-NLS-1$ //$NON-NLS-2$
		str = str.replace("Shift", "shift"); //$NON-NLS-1$ //$NON-NLS-2$
		str = str.replace("Alt", "alt"); //$NON-NLS-1$ //$NON-NLS-2$
		str = str.replace("\\", "BACK_SLASH"); //$NON-NLS-1$ //$NON-NLS-2$
		str = str.replace("Enter", "ENTER"); //$NON-NLS-1$ //$NON-NLS-2$
		str = str.replace("BkSp", "BACK_SPACE"); //$NON-NLS-1$ //$NON-NLS-2$
		str = str.replace("Ins", "INSERT"); //$NON-NLS-1$ //$NON-NLS-2$
		str = str.replace("Del", "DELETE"); //$NON-NLS-1$ //$NON-NLS-2$
		return KeyStroke.getKeyStroke(str);
	}

	private JMenuItem getMenuItem(MainFrame frame, String text,
			String keyStroke, String actionName) {
		JMenuItem item;
		Action action = null;
		if (actionName.equals(SUBMENU_KEYWORD)) {
			JMenu submenu = new JMenu();
			menuMap.put(text, submenu);
			item = submenu;
		} else {
			action = getAction(frame, actionName);
			item = action.createMenuItem();
		}
		item.setMnemonic(getMnemonic(text));
		item.setText(text.replaceAll(AMP_REGEX, EMPTY));
		if (!(item instanceof JMenu)) {
			item.setAccelerator(getKeyStroke(keyStroke));
		}
		if (action != null) {
			item.addActionListener(action);
		}
		return item;
	}

	private void addMenu(MainFrame frame, ICassavaTableModel model, int y) {
		String category = model.getValueAt(y, 0);
		if (category.length() == 0 || category.startsWith(IGNORE_KEYWORD)) {
			return;
		}
		String text = model.getValueAt(y, 1);
		String keyStroke = model.getValueAt(y, 2);
		String actionName = model.getValueAt(y, 3);
		if (category.equals(STARTUP_KEYWORD)) {
			frame.setStartUpAction(getAction(frame, actionName));
		} else if (category.equals(POPUP_KEYWORD)) {
			if (text.equals(SEPARATOR_KEYWORD)) {
				jPopupMenu.addSeparator();
			} else {
				JMenuItem item = getMenuItem(frame, text, keyStroke, actionName);
				jPopupMenu.add(item);
			}
		} else {
			if (text.equals(SEPARATOR_KEYWORD)) {
				getJMenu(category).addSeparator();
			} else {
				JMenuItem item = getMenuItem(frame, text, keyStroke, actionName);
				getJMenu(category).add(item);
			}
		}
	}

	public void readMenuCsv(MainFrame frame, JMenuBar jMenuBar,
			JPopupMenu jPopupMenu) {
		this.jMenuBar = jMenuBar;
		this.jPopupMenu = jPopupMenu;
		try {
			InputStream is;
			is = MenuBuilder.class.getResourceAsStream(USER_MENU_FILE_NAME);
			if (is == null) {
				is = MenuBuilder.class
						.getResourceAsStream(DEFAULT_MENU_FILE_NAME);
			}
			CsvReader reader = new CsvReader(DataFormat.RFC4180,
					new InputStreamReader(is, StandardCharsets.UTF_8));
			try {
				ICassavaTableModel model = reader.readTable();
				for (int y = 1; y <= model.getRowCount(); y++) {
					addMenu(frame, model, y);
				}
			} finally {
				reader.close();
			}
		} catch (IOException ex) {
			Log.error(ex);
		}
	}
}
