package net.asukaze.cassava.macro;

import net.asukaze.cassava.macro.parser.ASTArgument;
import net.asukaze.cassava.macro.parser.ASTFunctionDef;
import net.asukaze.cassava.macro.parser.ASTScript;
import net.asukaze.cassava.macro.parser.ASTStatement;
import net.asukaze.cassava.macro.parser.Node;
import net.asukaze.cassava.macro.parser.SimpleNode;

public class UserFunction {
	private String fileid;
	private String name;
	private String[] argNames;
	private SimpleNode body;

	public UserFunction(ASTFunctionDef node, String fileid) {
		int n = node.jjtGetNumChildren();
		int argCount = n - 1;
		String[] argNames = new String[argCount];
		for (int i = 0; i < argCount; i++) {
			ASTArgument arg = (ASTArgument) (node.jjtGetChild(i));
			argNames[i] = arg.nodeValue;
		}

		this.fileid = fileid;
		this.name = node.nodeValue;
		this.argNames = argNames;
		this.body = (SimpleNode) (node.jjtGetChild(n - 1));
	}

	public String getName() {
		return name;
	}

	public int getArgCount() {
		return argNames.length;
	}

	public Object run(Macro macro, Object[] args) {
		try {
			macro.pushEnvironment(fileid);
			Environment environment = macro.getEnvironment();
			for (int i = 0; i < argNames.length; i++) {
				Object value = environment.dereference(args[i]);
				environment.putVar(argNames[i], value);
			}
			return macro.accept(body);
		} catch (ReturnException ex) {
			return ex.getValue();
		} finally {
			macro.popEnvironment();
		}
	}

	public static Object run(FunctionCall call, Macro macro, boolean qualified) {
		String fileid;
		String funcname;
		if (qualified) {
			String[] names = call.getName().split("\\."); //$NON-NLS-1$
			fileid = names[0];
			funcname = names[1];
		} else {
			fileid = macro.getEnvironment().getCurrentFileId();
			funcname = call.getName();
		}
		int argCount = call.getArgs().length;
		UserFunction userFunction = getFunction(fileid, funcname, argCount,
				macro);
		if (userFunction == null) {
			return new FunctionNotFoundException(call);
		}
		return userFunction.run(macro, call.getArgs());
	}

	private static UserFunction getFunction(String fileid, String funcname,
			int argCount, Macro macro) {
		UserFunction userFunction = macro.getEnvironment().getFunction(fileid,
				funcname, argCount);
		if (userFunction != null) {
			return userFunction;
		}
		ASTScript script = macro.getScriptTree(fileid);
		if (script == null) {
			return null;
		}
		ASTFunctionDef funcNode = searchFunction(script, funcname, argCount);
		if (funcNode == null) {
			return null;
		}
		userFunction = new UserFunction(funcNode, fileid);
		macro.getEnvironment().putFunction(fileid, funcname, argCount,
				userFunction);
		return userFunction;
	}

	private static ASTFunctionDef searchFunction(ASTScript node,
			String funcname, int argCount) {
		int n = node.jjtGetNumChildren();
		for (int i = 0; i < n; i++) {
			ASTStatement child = (ASTStatement) node.jjtGetChild(i);
			ASTFunctionDef function = searchFunction(child, funcname, argCount);
			if (function != null) {
				return function;
			}
		}
		return null;
	}

	private static ASTFunctionDef searchFunction(ASTStatement node,
			String funcname, int argCount) {
		int n = node.jjtGetNumChildren();
		for (int i = 0; i < n; i++) {
			Node child = node.jjtGetChild(i);
			if (child instanceof ASTFunctionDef) {
				ASTFunctionDef function = (ASTFunctionDef) child;
				if (isFunction(function, funcname, argCount)) {
					return function;
				}
			}
		}
		return null;
	}

	private static boolean isFunction(ASTFunctionDef node, String funcname,
			int argCount) {
		int n = node.jjtGetNumChildren();
		if (funcname.equals(node.nodeValue) && argCount == (n - 1)) {
			return true;
		}
		return false;
	}
}
