package net.asukaze.cassava.macro.cui;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.StringReader;
import java.io.Writer;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import net.asukaze.cassava.core.model.CassavaViewModel;
import net.asukaze.cassava.core.model.DataFormat;
import net.asukaze.cassava.core.model.ICassavaTableModel;
import net.asukaze.cassava.macro.Macro;
import net.asukaze.cassava.macro.api.CUIMessageBoxFunctionHandler;
import net.asukaze.cassava.macro.api.ConsoleWriteFunctionHandler;
import net.asukaze.cassava.macro.api.ExportWriteFunctionHandler;
import net.asukaze.cassava.macro.parser.ASTScript;
import net.asukaze.cassava.macro.parser.CMSParser;
import net.asukaze.cassava.macro.parser.CMSParserConstants;
import net.asukaze.cassava.macro.parser.CMSParserTokenManager;
import net.asukaze.cassava.macro.parser.ParseException;
import net.asukaze.cassava.macro.parser.SimpleCharStream;
import net.asukaze.cassava.macro.parser.Token;
import net.asukaze.cassava.macro.parser.TokenMgrError;
import net.asukaze.cassava.macro.value.VarReference;

public class Interactive {
	private static final Set<String> CUIAPI_FUNCTIONS = new HashSet<String>(
			Arrays.asList(new String[] { "list", "save", "ls", "pwd" }));

	private ICassavaTableModel model;
	private BufferedReader in;
	private PrintStream out;
	private Writer writer;
	private InteractiveFunctionHandler cuiApi;

	public Interactive(ICassavaTableModel model, Writer writer,
			DataFormat dataFormat) {
		this.model = model;
		this.in = new BufferedReader(new InputStreamReader(System.in));
		this.out = System.out;
		this.writer = writer;
		this.cuiApi = new InteractiveFunctionHandler(in, out, dataFormat);
	}

	private Macro setupMacro() {
		Macro macro = new Macro();
		macro.setModel(model);
		macro.setViewModel(new CassavaViewModel());
		if (writer != null) {
			macro.registerAPIFunctionHandler(new ExportWriteFunctionHandler(
					writer));
		} else {
			macro.registerAPIFunctionHandler(new ConsoleWriteFunctionHandler(
					out));
		}
		macro.registerAPIFunctionHandler(new CUIMessageBoxFunctionHandler(in,
				out));
		macro.registerAPIFunctionHandler(cuiApi);
		macro.registerAPIFunctionHandler(new MenuEmulator(in, out));
		return macro;
	}

	private SimpleCharStream getSimpleCharStream(String string) {
		return new SimpleCharStream(new StringReader(string));
	}

	public void run(Macro macro) throws IOException {
		CMSParserTokenManager tokenManager = new CMSParserTokenManager(null);
		StringBuilder sb = new StringBuilder();
		while (true) {
			if (sb.length() == 0) {
				out.print("> ");
			} else {
				out.print("| ");
			}
			String line = in.readLine();
			if (line == null) {
				break;
			}
			sb.append(line);
			sb.append("\n");
			tokenManager.ReInit(getSimpleCharStream(sb.toString()));
			int blocklevel = 0;
			try {
				while (true) {
					Token token = tokenManager.getNextToken();
					if (token == null || token.kind == CMSParserConstants.EOF) {
						break;
					}
					if (token.kind == CMSParserConstants.LBLOCK) {
						blocklevel++;
					} else if (token.kind == CMSParserConstants.RBLOCK) {
						blocklevel--;
					}
				}
				if (blocklevel == 0) {
					sb.append(";");
					boolean exit = execute(sb.toString(), macro);
					if (exit) {
						return;
					}
					sb = new StringBuilder();
				}
			} catch (TokenMgrError err) {
				String message = err.getMessage();
				if (message.contains("Encountered: <EOF>")) {
					// šɂȂɂ΃p[Xł
				} else {
					System.err.println(message);
					sb = new StringBuilder();
				}
			} catch (Exception ex) {
				String message;
				if (ex instanceof ParseException) {
					message = ex.getMessage();
					message = message.replaceAll("\\.\\.\\.\\r?\\n\\s+", " ");
				} else {
					message = ex.getLocalizedMessage();
				}
				System.err.println(message);
				sb = new StringBuilder();
			}
		}

	}

	public void run() throws IOException {
		Macro macro = setupMacro();
		try {
			run(macro);
		} finally {
			macro.close();
		}
	}

	private boolean execute(String string, Macro macro) throws ParseException,
			IOException {
		string = string.trim();
		if (string.equals(";")) {
			return false;
		}
		CMSParser parser = new CMSParser(new StringReader(string));
		ASTScript script = parser.Script();
		int n = script.jjtGetNumChildren();
		for (int i = 0; i < n; i++) {
			Object value = macro.accept(script.jjtGetChild(i));
			if (value instanceof VarReference) {
				String cmd = ((VarReference) value).getName();
				if ("exit".equals(cmd) || "quit".equals(cmd)) {
					if (InteractiveFunctionHandler.confirmIfModified(model,
							"I܂H", in, out)) {
						return true;
					}
				} else if (CUIAPI_FUNCTIONS.contains(cmd)) {
					value = cuiApi.runAPIFunction(cmd, new Object[0], macro);
				}
			}
			if (value != null) {
				out.println(macro.toString(value));
			} else if (i == 0 || i < n - 1) {
				out.println("Ok.");
			}
		}
		return false;
	}

	public void close() throws IOException {
		in.close();
		out.close();
		if (writer != null) {
			writer.close();
		}
	}

	@Override
	protected void finalize() throws Throwable {
		close();
	}
}
