package net.asukaze.cassava.macro.cui;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;

import net.asukaze.cassava.core.model.CassavaTableModel;
import net.asukaze.cassava.core.model.CassavaViewModel;
import net.asukaze.cassava.core.model.DataFormat;
import net.asukaze.cassava.core.model.ICassavaTableModel;
import net.asukaze.cassava.core.util.DataFormatSetting;
import net.asukaze.cassava.io.util.CsvReader;
import net.asukaze.cassava.io.util.CsvWriter;
import net.asukaze.cassava.macro.APIFunctionHandler;
import net.asukaze.cassava.macro.Macro;
import net.asukaze.cassava.macro.MacroException;
import net.asukaze.cassava.macro.api.CUIMessageBoxFunctionHandler;
import net.asukaze.cassava.macro.api.ConsoleWriteFunctionHandler;
import net.asukaze.cassava.macro.api.ExportWriteFunctionHandler;
import net.asukaze.cassava.macro.api.TableWriteFunctionHandler;
import net.asukaze.cassava.macro.parser.ParseException;

public class Main {
	private static final String USAGE_FILENAME = "usage.txt"; //$NON-NLS-1$
	private static final String LINE_SEPARATOR = System
			.getProperty("line.separator"); //$NON-NLS-1$

	private InputStream script = null;
	private InputStream in = null;
	private String inputFileName = null;
	private OutputStream out = null;
	private String outputFileName = null;
	private boolean export = false;
	private DataFormat dataFormat = null;

	public Main(String[] args) throws ArgumentException, IOException {
		parseArgs(args);
		if (dataFormat == null) {
			dataFormat = estimateDataFormat();
		}
	}

	private DataFormat estimateDataFormat() {
		DataFormat result = null;
		DataFormatSetting setting = DataFormatSetting.INSTANCE;
		if (inputFileName != null) {
			result = setting.getDataFormatForFilename(inputFileName);
			if (result != null) {
				return result;
			}
		}
		if (outputFileName != null) {
			result = setting.getDataFormatForFilename(outputFileName);
			if (result != null) {
				return result;
			}
		}
		return setting.getDefault();
	}

	private ICassavaTableModel read() throws IOException {
		if (in == null) {
			in = System.in;
		}
		Reader inputReader = new InputStreamReader(in);
		CsvReader reader = new CsvReader(dataFormat, inputReader);
		try {
			return reader.readTable();
		} finally {
			reader.close();
			inputReader.close();
		}
	}

	private void write(ICassavaTableModel model) throws IOException {
		if (out == null) {
			out = System.out;
		}
		Writer outputWriter = new OutputStreamWriter(out);
		CsvWriter writer = new CsvWriter(dataFormat, outputWriter,
				LINE_SEPARATOR);
		try {
			writer.writeTable(model);
		} finally {
			writer.close();
			outputWriter.close();
		}
	}

	private void run() throws IOException, ParseException, ArgumentException {
		if (script != null) {
			runBatch();
		} else {
			runInteractive();
		}
	}

	private APIFunctionHandler getWriteFunctionHandler() {
		if (export) {
			if (out != null) {
				Writer writer = new OutputStreamWriter(out);
				return new ExportWriteFunctionHandler(writer);
			} else {
				return new ConsoleWriteFunctionHandler(System.out);
			}
		} else {
			return new TableWriteFunctionHandler();
		}
	}

	private void runBatch() throws IOException, ParseException {
		ICassavaTableModel model = read();
		Macro macro = new Macro();
		macro.setModel(model);
		macro.setViewModel(new CassavaViewModel());
		macro.registerAPIFunctionHandler(getWriteFunctionHandler());
		macro.registerAPIFunctionHandler(new CUIMessageBoxFunctionHandler(
				System.in, System.out));
		macro.registerAPIFunctionHandler(new MenuEmulator());
		macro.run(script);
		macro.close();
		if (!export) {
			write(model);
		}
	}

	private void runInteractive() throws IOException, ArgumentException {
		ICassavaTableModel model;
		if (in != null) {
			model = read();
		} else {
			model = new CassavaTableModel();
		}

		Writer writer;
		if (out != null) {
			writer = new OutputStreamWriter(out);
		} else {
			writer = null;
		}

		Interactive interactive = new Interactive(model, writer, dataFormat);
		try {
			interactive.run();
		} finally {
			interactive.close();
		}
	}

	private InputStream getFileInputStream(String filename)
			throws FileNotFoundException {
		if (filename == null || filename.equals("-")) {
			return System.in;
		} else {
			return new FileInputStream(filename);
		}
	}

	private OutputStream getFileOutputStream(String filename)
			throws FileNotFoundException {
		if (filename == null || filename.equals("-")) {
			return System.out;
		} else {
			return new FileOutputStream(filename);
		}
	}

	public void parseArgs(String[] args) throws ArgumentException, IOException {
		for (int i = 0; i < args.length; i++) {
			String arg = args[i];
			if (arg == null) {
				continue;
			}
			if (arg.startsWith("-")) { //$NON-NLS-1$
				if (arg.equals("-i") || arg.equals("--in")) { //$NON-NLS-1$ //$NON-NLS-2$
					if (in == null && (i + 1) < args.length) {
						inputFileName = args[++i];
						in = getFileInputStream(inputFileName);
					} else {
						throw new ArgumentException();
					}
				} else if (arg.equals("-o") || arg.equals("--out")) { //$NON-NLS-1$ //$NON-NLS-2$
					if (out == null && (i + 1) < args.length) {
						outputFileName = args[++i];
						out = getFileOutputStream(outputFileName);
					} else {
						throw new ArgumentException();
					}
				} else if (arg.equals("-x") || arg.equals("--export")) { //$NON-NLS-1$ //$NON-NLS-2$
					export = true;
				} else if (arg.equals("-xo") || arg.equals("-ox")) { //$NON-NLS-1$ //$NON-NLS-2$
					export = true;
					if (out == null && (i + 1) < args.length) {
						outputFileName = args[++i];
						out = getFileOutputStream(outputFileName);
					} else {
						throw new ArgumentException();
					}
				} else if (arg.equals("-f") || arg.equals("--format")) { //$NON-NLS-1$ //$NON-NLS-2$
					if ((i + 1) < args.length) {
						dataFormat = DataFormatSetting.INSTANCE.get(args[++i]);
						if (dataFormat == null) {
							throw new MacroException("Format Not Found: "
									+ args[i]);
						}
					} else {
						throw new ArgumentException();
					}
				} else if (arg.equals("-v") || arg.equals("--version")) { //$NON-NLS-1$ //$NON-NLS-2$
					showVersion();
					throw new ArgumentException(false);
				} else if (arg.equals("-")) { //$NON-NLS-1$
					if (script == null) {
						script = System.in;
					} else {
						throw new ArgumentException();
					}
				} else {
					throw new ArgumentException();
				}
			} else {
				if (script == null) {
					script = getFileInputStream(arg);
				} else {
					throw new ArgumentException();
				}
			}
		}
	}

	private static void print(InputStream inputStream) throws IOException {
		BufferedReader br = new BufferedReader(new InputStreamReader(
				inputStream));
		try {
			String line;
			while ((line = br.readLine()) != null) {
				System.out.println(line);
			}
		} finally {
			br.close();
			inputStream.close();
		}
	}

	private void showVersion() throws IOException {
		print(Main.class.getResourceAsStream("/cuissava-release"));
	}

	public static void main(String[] args) throws IOException, ParseException {
		try {
			new Main(args).run();
		} catch (ArgumentException ex) {
			if (ex.getShowHelp()) {
				print(Main.class.getResourceAsStream(USAGE_FILENAME));
				System.out.println();
			}
		} catch (Exception ex) {
			String message = ex.getLocalizedMessage();
			if (message != null && message.length() > 0) {
				System.err.println(message);
			} else {
				ex.printStackTrace();
			}
		}
	}
}
