package net.asukaze.cassava.macro;

import static org.junit.Assert.*;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.StringWriter;

import net.asukaze.cassava.core.model.CassavaTableModel;
import net.asukaze.cassava.core.model.ICassavaTableModel;
import net.asukaze.cassava.macro.api.ExportWriteFunctionHandler;
import net.asukaze.cassava.macro.parser.ParseException;

import org.junit.Before;
import org.junit.Test;

public class MacroTest {
	private static final String NL = "\r\n"; //$NON-NLS-1$
	private Macro macro;

	@Before
	public void setUp() {
		macro = new Macro();
	}

	public String run(String input) throws ParseException, IOException {
		StringWriter sw = new StringWriter();
		macro.registerAPIFunctionHandler(new ExportWriteFunctionHandler(sw));
		macro.run(new ByteArrayInputStream(input.getBytes()));
		return sw.toString();
	}

	public Object runExpr(String input) throws ParseException, IOException {
		return macro.runExpr(new ByteArrayInputStream(input.getBytes()));
	}

	@Test
	public void test2() throws ParseException, IOException {
		String result = run("writeln(\"1helloworld\"); writeln(\"helloworld2\");"); //$NON-NLS-1$
		assertEquals("1helloworld" + NL + "helloworld2" + NL, result); //$NON-NLS-1$ //$NON-NLS-2$
	}

	@Test
	public void test3() throws ParseException, IOException {
		String result = run("writeln(3); writeln(4);"); //$NON-NLS-1$
		assertEquals("3" + NL + "4" + NL, result); //$NON-NLS-1$ //$NON-NLS-2$
	}

	@Test
	public void test4() throws ParseException, IOException {
		String result = run(""); //$NON-NLS-1$
		assertEquals("", result); //$NON-NLS-1$
	}

	@Test
	public void test5() throws ParseException, IOException {
		String result = run("writeln(5+6); writeln(7+8);"); //$NON-NLS-1$
		assertEquals("11" + NL + "15" + NL, result); //$NON-NLS-1$ //$NON-NLS-2$
	}

	@Test
	public void test6() throws ParseException, IOException {
		String result = run("writeln(9-11); writeln(10-13);"); //$NON-NLS-1$
		assertEquals("-2" + NL + "-3" + NL, result); //$NON-NLS-1$ //$NON-NLS-2$
	}

	@Test
	public void test7() throws ParseException, IOException {
		String result = run("writeln(11-9+1); writeln(12-9+1);"); //$NON-NLS-1$
		assertEquals("3" + NL + "4" + NL, result); //$NON-NLS-1$ //$NON-NLS-2$
	}

	@Test
	public void test8() throws ParseException, IOException {
		String result = run("writeln(13+5*3-(2+1));"); //$NON-NLS-1$
		assertEquals("25" + NL, result); //$NON-NLS-1$
	}

	@Test
	public void test9() throws ParseException, IOException {
		String result = run("writeln(\"ʂ\" + (13+5*3-(2+1)));"); //$NON-NLS-1$
		assertEquals("ʂ25" + NL, result); //$NON-NLS-1$
	}

	@Test
	public void test10() throws ParseException, IOException {
		String result = run("writeln(\"ʂ\" + 4 + 3);"); //$NON-NLS-1$
		assertEquals("ʂ43" + NL, result); //$NON-NLS-1$
	}

	@Test
	public void test11() throws ParseException, IOException {
		String result = run("writeln(\"ʂ\" + (4 + 3));"); //$NON-NLS-1$
		assertEquals("ʂ7" + NL, result); //$NON-NLS-1$
	}

	@Test
	public void test12() throws ParseException, IOException {
		String result = run("writeln(2.1 + 4.3);"); //$NON-NLS-1$
		assertEquals("6.4" + NL, result); //$NON-NLS-1$
	}

	@Test
	public void test13() throws ParseException, IOException {
		String result = run("writeln(6.8 / 2);"); //$NON-NLS-1$
		assertEquals("3.4" + NL, result); //$NON-NLS-1$
	}

	@Test
	public void test14() throws ParseException, IOException {
		String result = run("writeln(\"hello, \\\"\\n\\t\\\\\");"); //$NON-NLS-1$
		assertEquals("hello, \"\n\t\\" + NL, result); //$NON-NLS-1$
	}

	@Test
	public void test15() throws ParseException, IOException {
		String result = run("writeln(\"\\x41\\x62\\x35\\x2a\\x4F\");"); //$NON-NLS-1$
		assertEquals("Ab5*O" + NL, result); //$NON-NLS-1$
	}

	@Test
	public void test16() throws ParseException, IOException {
		String result = run("abc=\"hello\";"); //$NON-NLS-1$
		assertEquals("", result); //$NON-NLS-1$
		assertEquals("hello", macro.getVar("abc")); //$NON-NLS-1$ //$NON-NLS-2$
	}

	@Test
	public void test17() throws ParseException, IOException {
		String result = run("abc1xyz23 = 345 + 3.7;"); //$NON-NLS-1$
		assertEquals("", result); //$NON-NLS-1$
		assertEquals(348.7, macro.getVar("abc1xyz23")); //$NON-NLS-1$
	}

	@Test
	public void test18() throws ParseException, IOException {
		String result = run("a = 3; write(\"Hello, \"); writeln(\"World!\"); writeln(a);"); //$NON-NLS-1$
		assertEquals("Hello, World!" + NL + "3" + NL, result); //$NON-NLS-1$ //$NON-NLS-2$
		assertEquals(3.0, macro.getVar("a")); //$NON-NLS-1$
	}

	@Test
	public void test19() throws ParseException, IOException {
		ICassavaTableModel model = new CassavaTableModel();
		model.setValueAt("ABC", 0, 0); //$NON-NLS-1$
		model.setValueAt("DEF", 0, 1); //$NON-NLS-1$
		macro.setModel(model);
		String result = run("aa = [1,1]; [3,4] = aa + [2,1];"); //$NON-NLS-1$
		assertEquals("", result); //$NON-NLS-1$
		assertEquals("ABC", macro.getVar("aa")); //$NON-NLS-1$ //$NON-NLS-2$
		assertEquals("ABCDEF", model.getValueAt(3, 2)); //$NON-NLS-1$
	}

	@Test
	public void test20() throws ParseException, IOException {
		ICassavaTableModel model = new CassavaTableModel();
		model.setValueAt("ABC", 0, 0); //$NON-NLS-1$
		model.setValueAt("DEF", 0, 1); //$NON-NLS-1$
		macro.setModel(model);
		String result = run("aa = cell(1,1); cell(3,4) = aa + cell(2,1);"); //$NON-NLS-1$
		assertEquals("", result); //$NON-NLS-1$
		assertEquals("ABC", macro.getVar("aa")); //$NON-NLS-1$ //$NON-NLS-2$
		assertEquals("ABCDEF", model.getValueAt(3, 2)); //$NON-NLS-1$
	}

	@Test
	public void test21() throws ParseException, IOException {
		String result = run("foo = 0; if(foo){ bar = \"ng\"; }"); //$NON-NLS-1$
		assertEquals("", result); //$NON-NLS-1$
		assertEquals(null, macro.getVar("bar")); //$NON-NLS-1$ //$NON-NLS-2$
	}

	@Test
	public void test22() throws ParseException, IOException {
		String result = run("foo = 1; if(foo){ bar = \"ok\"; }"); //$NON-NLS-1$
		assertEquals("", result); //$NON-NLS-1$
		assertEquals("ok", macro.getVar("bar")); //$NON-NLS-1$ //$NON-NLS-2$
	}

	@Test
	public void test23() throws ParseException, IOException {
		String result = run("foo = 0; if(foo){ bar = \"ngg\"; }else{ bar = \"okk\"; }"); //$NON-NLS-1$
		assertEquals("", result); //$NON-NLS-1$
		assertEquals("okk", macro.getVar("bar")); //$NON-NLS-1$ //$NON-NLS-2$
	}

	@Test
	public void test24() throws ParseException, IOException {
		String result = run("foo = 1; if(foo){ bar = \"okkk\"; }else{ bar = \"nggg\"; };"); //$NON-NLS-1$
		assertEquals("", result); //$NON-NLS-1$
		assertEquals("okkk", macro.getVar("bar")); //$NON-NLS-1$ //$NON-NLS-2$
	}

	@Test
	public void test25() throws ParseException, IOException {
		String result = run("x=0; y=1; if(x){bar = 1;}else if(y){bar = 2;}else{bar=3;}"); //$NON-NLS-1$
		assertEquals("", result); //$NON-NLS-1$
		assertEquals(2.0, macro.getVar("bar")); //$NON-NLS-1$
	}

	@Test
	public void test26() throws ParseException, IOException {
		String result = run("n=10; s=0; while(n){ s=s+n; n=n-1; } writeln(s);"); //$NON-NLS-1$
		assertEquals("55" + NL, result); //$NON-NLS-1$
		assertEquals(0.0, macro.getVar("n")); //$NON-NLS-1$
		assertEquals(55.0, macro.getVar("s")); //$NON-NLS-1$
	}

	@Test
	public void test27() throws ParseException, IOException {
		String result = run("s=0; for(n=10; n; n=n-1){ s=s+n; write(n); } writeln(s);"); //$NON-NLS-1$
		assertEquals("1098765432155" + NL, result); //$NON-NLS-1$
		assertEquals(0.0, macro.getVar("n")); //$NON-NLS-1$
		assertEquals(55.0, macro.getVar("s")); //$NON-NLS-1$
	}

	@Test
	public void test28() throws ParseException, IOException {
		String result = run("s=0; For(n to 10){ s=s+n; write(n); } writeln(s);"); //$NON-NLS-1$
		assertEquals("1234567891055" + NL, result); //$NON-NLS-1$
		assertEquals(10, macro.getVar("n")); //$NON-NLS-1$
		assertEquals(55.0, macro.getVar("s")); //$NON-NLS-1$
	}

	@Test(expected = ParseException.class)
	public void test29() throws ParseException, IOException {
		run("s=0; For(n too 10){ s=s+n; write(n); } writeln(s);"); //$NON-NLS-1$
	}

	@Test
	public void test30() throws ParseException, IOException {
		String result = run("s=0;\r\n For(n to 10)\n" //$NON-NLS-1$
				+ "{ s=/* */s+/***/n/*******/; // comment \n" //$NON-NLS-1$
				+ " write/* hohoho */ (n); }/////\r" //$NON-NLS-1$
				+ "writeln/*//* //* //****/(s);"); //$NON-NLS-1$
		assertEquals("1234567891055" + NL, result); //$NON-NLS-1$
		assertEquals(10, macro.getVar("n")); //$NON-NLS-1$
		assertEquals(55.0, macro.getVar("s")); //$NON-NLS-1$
	}

	@Test
	public void test31() throws ParseException, IOException {
		String result = run("writeln(12); sample1.func1(); writeln(34); sample1.func1(5,6,7); writeln(89);"); //$NON-NLS-1$
		assertEquals("12" + NL + "sample1.func1/0" + NL + "34" + NL //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
				+ "sample1.func1/3:5,6,7" + NL + "89" + NL, result); //$NON-NLS-1$ //$NON-NLS-2$
	}

	@Test
	public void test32() throws ParseException, IOException {
		String result = run("writeln(sample2.func1()); writeln(sample2.func1(2,3,4)); writeln(sample2.func1(5,\"6\",7));"); //$NON-NLS-1$
		assertEquals("144" + NL + "9" + NL + "567" + NL, result); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
	}

	@Test
	public void test33() throws ParseException, IOException {
		Object result = runExpr("int(\"\")"); //$NON-NLS-1$
		assertEquals(Integer.valueOf(0), result);
	}

	@Test(expected = FunctionNotFoundException.class)
	public void test34() throws ParseException, IOException {
		runExpr("notdefinedfunction()"); //$NON-NLS-1$
	}

	@Test
	public void test35() throws ParseException, IOException {
		Object result = runExpr("dont_define_this + 123 + 456"); //$NON-NLS-1$
		assertEquals(579.0, result);
	}

	@Test
	public void test36() throws ParseException, IOException {
		Object result = runExpr("123 + \"456\""); //$NON-NLS-1$
		assertEquals("123456", result);
	}

	@Test
	public void test37() throws ParseException, IOException {
		Object result = runExpr("123 + \"\" + 456"); //$NON-NLS-1$
		assertEquals("123456", result);
	}

	@Test
	public void test38() throws ParseException, IOException {
		Object result = runExpr("Null"); //$NON-NLS-1$
		assertNull(macro.dereference(result));
	}

	@Test
	public void test39() throws ParseException, IOException {
		Object result = runExpr("3 + Null + 4"); //$NON-NLS-1$
		assertEquals("7", macro.toString(result));
	}

  @Test
  public void testBigDouble_outputsPlainString() throws ParseException, IOException {
    Object result = runExpr("012345679.0 - 1.0"); //$NON-NLS-1$
    assertEquals("12345678", macro.toString(result));
  }

  @Test
  public void testBigDoubleCell_outputsOriginalString() throws ParseException, IOException {
    ICassavaTableModel model = new CassavaTableModel();
    model.setValueAt("012345678.0", 0, 0); //$NON-NLS-1$
    macro.setModel(model);
    Object result = runExpr("[1,1]"); //$NON-NLS-1$
    assertEquals("012345678.0", macro.toString(result));
  }
}
